@@ -5,23 +5,32 @@ use warnings;
 package Log::Any::Proxy;
 
 # ABSTRACT: Log::Any generator proxy object
-our $VERSION = '1.045';
+our $VERSION = '1.705';
 
 use Log::Any::Adapter::Util ();
 use overload;
 
+sub _stringify_params {
+    my @params = @_;
+
+    return map {
+        !defined($_)
+          ? '<undef>'
+          : ref($_) ? (
+            overload::OverloadedStringify($_)
+            ? "$_"
+            : Log::Any::Adapter::Util::dump_one_line($_)
+          )
+          : $_
+    } @params;
+}
+
 sub _default_formatter {
     my ( $cat, $lvl, $format, @params ) = @_;
     return $format->() if ref($format) eq 'CODE';
-    my @new_params =
-      map {
-           !defined($_) ? '<undef>'
-          : ref($_)     ? (
-            overload::OverloadedStringify($_) ? "$_"
-          : Log::Any::Adapter::Util::dump_one_line($_)
-        )
-          : $_
-      } @params;
+
+    my @new_params = _stringify_params(@params);
+
     # Perl 5.22 adds a 'redundant' warning if the number parameters exceeds
     # the number of sprintf placeholders.  If a user does this, the warning
     # is issued from here, which isn't very helpful.  Doing something
@@ -40,7 +49,11 @@ sub new {
     }
     unless ( $self->{category} ) {
         require Carp;
-        Carp::croak("$class requires an 'category' parameter")
+        Carp::croak("$class requires a 'category' parameter");
+    }
+    unless ( $self->{context} ) {
+        require Carp;
+        Carp::croak("$class requires a 'context' parameter");
     }
     bless $self, $class;
     $self->init(@_);
@@ -54,7 +67,7 @@ sub clone {
 
 sub init { }
 
-for my $attr (qw/adapter filter formatter prefix/) {
+for my $attr (qw/adapter filter formatter prefix context/) {
     no strict 'refs';
     *{$attr} = sub { return $_[0]->{$attr} };
 }
@@ -76,9 +89,28 @@ foreach my $name ( Log::Any::Adapter::Util::logging_methods(), keys(%aliases) )
     };
     *{$name} = sub {
         my ( $self, @parts ) = @_;
-        my $message = join(" ", grep { defined($_) && length($_) } @parts );
-        if ( length $message ) {
-            $message = $self->{filter}->( $self->{category}, $numeric, $message )
+        return if !$self->{adapter}->$is_realname && !defined wantarray;
+
+        my $structured_logging =
+            $self->{adapter}->can('structured') && !$self->{filter};
+
+        if ($structured_logging) {
+            unshift @parts, $self->{prefix} if $self->{prefix};
+            $self->{adapter}
+              ->structured( $realname, $self->{category}, @parts, grep { scalar keys %$_ } $self->{context});
+            return unless defined wantarray;
+        }
+
+        @parts = grep { defined($_) && length($_) } @parts;
+
+        # last part might be a hashref - if so, stringify
+        push @parts, _stringify_params(pop @parts) if ( @parts && ((ref $parts[-1] || '') eq ref {}));
+
+        push @parts, _stringify_params($self->{context}) if %{$self->{context}};
+        my $message = join( " ", @parts );
+        if ( length $message && !$structured_logging ) {
+            $message =
+              $self->{filter}->( $self->{category}, $numeric, $message )
               if defined $self->{filter};
             if ( defined $message and length $message ) {
                 $message = "$self->{prefix}$message"
@@ -90,7 +122,7 @@ foreach my $name ( Log::Any::Adapter::Util::logging_methods(), keys(%aliases) )
     };
     *{$namef} = sub {
         my ( $self, @args ) = @_;
-        return unless $self->{adapter}->$is_realname;
+        return if !$self->{adapter}->$is_realname && !defined wantarray;
         my $message =
           $self->{formatter}->( $self->{category}, $numeric, @args );
         return unless defined $message and length $message;
@@ -115,7 +147,7 @@ Log::Any::Proxy - Log::Any generator proxy object
 
 =head1 VERSION
 
-version 1.045
+version 1.705
 
 =head1 SYNOPSIS
 
@@ -169,6 +201,9 @@ logged.  Otherwise, the return value is passed to the logging adapter.
 Numeric levels range from 0 (emergency) to 8 (trace).  Constant functions
 for these levels are available from L<Log::Any::Adapter::Util>.
 
+Configuring a filter disables structured logging, even if the
+configured adapter supports it.
+
 =head2 formatter
 
 A code reference to format messages given to the C<*f> methods (C<tracef>,
@@ -261,6 +296,54 @@ string is then returned.
 Numeric levels range from 0 (emergency) to 8 (trace).  Constant functions
 for these levels are available from L<Log::Any::Adapter::Util>.
 
+=head2 Logging Structured Data
+
+If you have data in addition to the text you want to log, you can
+specify a hashref after your string. If the configured adapter
+supports structured data, it will receive the hashref as-is, otherwise
+it will be converted to a string using L<Data::Dumper> and will be
+appended to your text.
+
+=head1 TIPS
+
+=head2 UTF-8 in Data Structures
+
+If you have high-bit characters in a data structure being passed to a log
+method, Log::Any will output that data structure with the high-bit
+characters encoded as C<\x{###}>, Perl's escape sequence for high-bit
+characters. This is because the L<Data::Dumper> module escapes those
+characters.
+
+    use utf8;
+    use Log::Any qw( $log );
+    my @data = ( "Привет мир" ); # Hello, World!
+    $log->infof("Got: %s", \@data);
+    # Got: ["\x{41f}\x{440}\x{438}\x{432}\x{435}\x{442} \x{43c}\x{438}\x{440}"]
+
+If you want to instead display the actual characters in your log file or
+terminal, you can use the L<Data::Dumper::AutoEncode> module. To wire this
+up into Log::Any, you must pass a custom C<formatter> sub:
+
+    use utf8;
+    use Data::Dumper::AutoEncode;
+
+    sub log_formatter {
+        my ( $category, $level, $format, @params ) = @_;
+        # Run references through Data::Dumper::AutoEncode
+        @params = map { ref $_ ? eDumper( $_ ) : $_ } @params;
+        return sprintf $format, @params;
+    }
+
+    use Log::Any '$log', formatter => \&log_formatter;
+
+This formatter changes the output to:
+
+	Got: $VAR1 = [
+			  'Привет мир'
+			];
+
+Thanks to L<@denis-it|https://github.com/denis-it> for this tip!
+
 =head1 AUTHORS
 
 =over 4
@@ -277,11 +360,19 @@ David Golden <dagolden@cpan.org>
 
 Doug Bell <preaction@cpan.org>
 
+=item *
+
+Daniel Pittman <daniel@rimspace.net>
+
+=item *
+
+Stephen Thirlwall <sdt@cpan.org>
+
 =back
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2016 by Jonathan Swartz, David Golden, and Doug Bell.
+This software is copyright (c) 2017 by Jonathan Swartz, David Golden, and Doug Bell.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
